import datetime


def windows_timestamp_to_datetime(ts: float) -> datetime.datetime:
    """
    Windows timestamp conversion is bugged on windows : https://github.com/python/cpython/issues/81708
    format : https://devblogs.microsoft.com/oldnewthing/20090306-00/?p=18913

    :param ts: timestamp windows filetime 100-nanosecond intervals since January 1, 1601
    :return:
    """
    windows_epoch = datetime.datetime(1601, 1, 1, tzinfo=datetime.timezone.utc)
    return windows_epoch + datetime.timedelta(microseconds=ts / 10)



"""
Files containing type containing each section
"""

import collections
import io
import json
import logging
import struct
import uuid
from enum import Enum
from typing import BinaryIO, Optional, Any

from wddh.utils import windows_timestamp_to_datetime

logger = logging.getLogger(__name__)


def strip_trailing_0_byte_wstr(b: bytes) -> bytes:
    if b.endswith(b"\x00\x00"):
        return b[:-2]
    else:
        return b


class TLVTypeEnum(Enum):
    UNKNOWN_0 = 0x0  # Maybe bool?
    UNKNOWN_1 = 0x1  # Only found in Blob
    SINT32 = 0x05
    UINT32 = 0x06
    UINT64 = 0x08
    FILETIME = 0x0A
    WSTRING = 0x15
    UUID = 0x1E
    BLOB = 0x28
    MISSING = -1

    @classmethod
    def _missing_(cls, value):
        """
        Set default value to prevent crash is value is missing
        Args:
            value:

        Returns:

        """
        return cls.MISSING


class LTVEntry:
    """
    Represent a Length Type Value entry
    """

    def __init__(
        self,
        fh: BinaryIO,
        expected_type: Optional[TLVTypeEnum] = None,
        expected_len: Optional[int] = None,
    ):
        """
        Create a TLV entry from a BinaryIO stream
        read empty byte after data in order to have an 8 bytes alignement
        log an info is type is not expected type

        Args:
            fh:
            expected_type:
            expected_len:
        """
        self.size: int = struct.unpack("<l", fh.read(4))[0]
        self.type_as_int = struct.unpack("<l", fh.read(4))[0]
        self.type: TLVTypeEnum = TLVTypeEnum(self.type_as_int)
        if expected_type and self.type != expected_type:
            logger.warning(
                "Do not find expected type at offset 0x%x. Expected : %s, Found : %s",
                fh.tell() - 4,
                expected_type,
                self.type,
            )
        self.raw_data: bytes = fh.read(self.size)
        # 8 byte alignement for the whole struct is expected
        padding = self.size % 8
        if padding:
            fh.read(8 - padding)

    def __repr__(self):
        return f"<L:{self.size}|T:{self.type_as_int}|V:{self.raw_data}>"


class U64Entry(LTVEntry):
    def __init__(self, fh: BinaryIO):
        super().__init__(fh, expected_type=TLVTypeEnum.UINT64, expected_len=8)
        self.as_u64 = struct.unpack("<Q", self.raw_data)[0]


class U32Entry(LTVEntry):
    def __init__(self, fh: BinaryIO):
        super().__init__(fh, expected_type=TLVTypeEnum.UINT32, expected_len=4)
        self.as_u32 = struct.unpack("<L", self.raw_data)[0]


class S32Entry(LTVEntry):
    """
    Signed 32 bytes
    """

    def __init__(self, fh: BinaryIO):
        super().__init__(fh, expected_type=TLVTypeEnum.SINT32, expected_len=4)
        self.as_s32 = struct.unpack("<l", self.raw_data)[0]


class UuidEntry(LTVEntry):
    def __init__(self, fh: BinaryIO):
        super().__init__(fh, expected_type=TLVTypeEnum.UUID, expected_len=0x10)
        self.as_uuid = str(uuid.UUID(bytes_le=self.raw_data))


class WStrEntry(LTVEntry):
    def __init__(self, fh: BinaryIO, errors="backslashreplace"):
        """

        Args:
            fh:
            errors: decoding errors handling mode (backslashreplace, errors etc...)
        """
        super().__init__(fh, expected_type=TLVTypeEnum.WSTRING)

        self.as_str = self.raw_data.decode("utf16", errors=errors)
        if self.as_str.endswith("\x00"):
            self.as_str = self.as_str[:-1]


class FiletimeEntry(LTVEntry):
    """
    Represent en windows filetime entry

    References:
        https://learn.microsoft.com/en-us/windows/win32/api/minwinbase/ns-minwinbase-filetime
    """

    def __init__(self, fh: BinaryIO):
        super().__init__(fh, expected_type=TLVTypeEnum.FILETIME, expected_len=8)
        self.as_u64 = struct.unpack("<Q", self.raw_data)[0]
        self.as_datetime = windows_timestamp_to_datetime(float(self.as_u64))

    def __str__(self):
        return self.as_datetime.isoformat()


class ProbablyBoolEntry(LTVEntry):
    def __init__(self, fh: BinaryIO):
        super().__init__(fh, expected_type=TLVTypeEnum.UNKNOWN_0, expected_len=4)
        self.as_u32 = struct.unpack("<L", self.raw_data)[0]


class BlobEnumType(Enum):
    UINT32 = 0x03  # Bool
    INT64 = 0x04  # time or int?
    BOOL = 0x05  # time or int?
    WSTRING = 0x06
    STRING = 0x07  # ASCII string


class BlobEntry(LTVEntry, collections.abc.Mapping):
    def __init__(self, fh: BinaryIO, errors="backslashreplace"):
        """

        Args:
            fh:
            errors: decoding errors handling mode (backslashreplace, errors etc...)
        """
        fh_start = fh.tell()
        super().__init__(fh, expected_type=TLVTypeEnum.BLOB)
        section_fh = io.BytesIO(self.raw_data)

        self.blob_map: dict[str, str | int | bool] = {}
        # May be empty
        if self.size == 0:
            return
        unknown_section = struct.unpack(
            "<l", section_fh.read(0x04)
        )[
            0
        ]  # These byte seems to contain unused information like repetition of blob length,
        # Look like a specific type
        if unknown_section == self.size:
            # first 0x04 byte are the blob length
            # Only found in the same samples were magic were absent from header
            pass
        else:
            # On most sample, 0x18 byte section with unknown information
            section_fh.read(0x14)
        while section_fh.tell() < self.size:
            value_type_int = None
            try:
                readable_key, value = self.read_key_val_entry(section_fh)
                if readable_key in self.blob_map:
                    logger.warning(
                        "Key already existing : %s, overriding previous key",
                        readable_key,
                    )
                self.blob_map[readable_key] = value
                value_type_int = None
            except struct.error:
                logger.warning(
                    "Fail to fully parse blob starting at offset 0x%x, current offset : 0x%x. Type : %s. Blob length :%d",
                    fh_start,
                    fh_start + section_fh.tell(),
                    value_type_int,
                    self.size,
                )

    def read_key_val_entry(self, section_fh) -> tuple[str, Any]:
        """
        Blob is made of kind of a key value list
        :param section_fh:
        :return: [key, value]
        """

        readable_key = self.read_w_str(section_fh)
        value_type_int = self.read_uint32(section_fh)
        logger.debug(
            "[BLOB]Reading for type %s for key %s", value_type_int, readable_key
        )

        match value_type_int:
            case BlobEnumType.WSTRING.value:
                return readable_key, self.read_w_str(section_fh)
            case BlobEnumType.UINT32.value:
                return readable_key, self.read_uint32(section_fh)
            case BlobEnumType.BOOL.value:
                content = section_fh.read(1)
                return readable_key, struct.unpack("?", content)[0]
            case BlobEnumType.INT64.value:
                content = section_fh.read(8)
                return readable_key, struct.unpack("<Q", content)[0]
            case BlobEnumType.STRING.value:
                return readable_key, self.read_str(section_fh)
            case _:
                logger.info(
                    "unknown blog blob TLV type : %d, will probably trying to process it as a string",
                    value_type_int,
                )
                return readable_key, self.read_w_str(section_fh)

    def __repr__(self):
        return json.dumps(self.blob_map)

    def read_w_str(self, section_fh: BinaryIO) -> str:
        value_len = self.read(section_fh, 4, "<L")
        content = strip_trailing_0_byte_wstr(section_fh.read(value_len))
        return content.decode("utf-16", errors="backslashreplace")

    def read_str(self, section_fh: BinaryIO) -> str:
        value_len = self.read(section_fh, 4, "<L")
        content = strip_trailing_0_byte_wstr(section_fh.read(value_len))
        return content.decode("ascii", errors="backslashreplace")

    def read_uint32(self, section_fh: BinaryIO) -> int:
        return self.read(section_fh, 4, "<L")

    def read_uint64(self, section_fh: BinaryIO) -> int:
        return self.read(section_fh, 8, "<Q")

    def read_bool(self, section_fh: BinaryIO) -> int:
        _ = section_fh.read(1)
        return self.read(section_fh, 1, "<?")

    def read(self, section_fh: BinaryIO, size: int, struct_fmt: str):
        """
        Read size byte from section fh and return first part of the unpacked struct using struct_fmt
        Args:
            section_fh:
            size:
            struct_fmt:

        Returns:

        """
        content = section_fh.read(size)  # todo ensure readed size == expected size
        return struct.unpack(struct_fmt, content)[0]

    def __getitem__(self, key):
        return self.blob_map[key]

    def __iter__(self):
        yield from self.blob_map.items()

    def __len__(self):
        return len(self.blob_map)

    def keys(self):
        return list(self.blob_map.keys())



import logging
from functools import reduce
from typing import BinaryIO, Optional

from wddh.wddh_part import (
    WDDHFlags,
    WDDHFooter,
    WDDHHeader,
    WDDHInformation,
    WDDHMetadata,
    WDDHMetadata2,
    WDDHOptionalSection,
    WDDHPart,
)

logger = logging.getLogger(__name__)


class WDDHClean(WDDHPart):
    """
    THis class is used to parse a windows defender detection History in the cleaniest way possible and will fail if file is incomplete or corrupted
    """

    def __init__(self, fh: BinaryIO, offset: Optional[int] = None):
        if offset:
            fh.seek(offset)
        logger.debug("Processing header at offset 0x%x", fh.tell())
        self.header = WDDHHeader(fh)
        logger.debug("Processing flags at offset 0x%x", fh.tell())
        self.flag_section = WDDHFlags(fh)
        self.alert_details = []
        for _ in range(self.flag_section.alert_detail_count):
            self.alert_details.append(WDDHInformation(fh))
        self.metadata = WDDHMetadata(fh)
        self.optional = None
        if self.flag_section.flag_1 == 4:
            self.optional = WDDHOptionalSection(fh)
        self.metadata_2 = WDDHMetadata2(fh)
        self.alert_details_2 = []
        for _ in range(self.metadata_2.count_following_information_section):
            self.alert_details_2.append(WDDHInformation(fh))
        self.footer = WDDHFooter(fh)

    def __repr__(self):
        return str(self.__dict__)

    def as_short(self) -> dict:
        """

        Returns: Only essential information, as dict

        """
        return {
            "threat_id": self.header.threat_id,
            "threat_name": self.header.threat_name,
            "threat_status": self.flag_section.threat_status_id.name,
            "domain_user": self.metadata_2.domain_user,
            "domain_user_group": self.metadata_2.domain_user_group,
            "process_name": self.metadata_2.process_name,
            "initial_detection_time": self.metadata_2.initial_detection_time,
            "remediation": self.metadata_2.remediation,
            "ressources": [
                f"{r.ressource_type} {r.ressource_location}" for r in self.alert_details
            ],
            "misc": reduce(
                lambda a, b: dict(a, **b), list(x.blob for x in self.alert_details), {}
            ),
        }



"""
Files containing class representing each section of the Windows Defender detection history files
"""

import collections
import json
import logging
from abc import ABC
from typing import TYPE_CHECKING, BinaryIO

from wddh.ltv_reader import (
    BlobEntry,
    FiletimeEntry,
    ProbablyBoolEntry,
    S32Entry,
    U32Entry,
    U64Entry,
    UuidEntry,
    WStrEntry,
)
from wddh.mp_threat_dectection_enum import ThreatStatusID

if TYPE_CHECKING:
    pass

logger = logging.getLogger(__name__)


class WDDHPart(collections.abc.Mapping, ABC):
    def __repr__(self):
        return json.dumps(self.__dict__, default=str)

    def keys(self):
        return list(self.__dict__.keys())

    def __getitem__(self, key):
        return self.__dict__[key]

    def __iter__(self):
        yield from self.__dict__.items()

    def __len__(self):
        return len(self.__dict__)


class WDDHHeader(WDDHPart):
    def __init__(self, fh: BinaryIO, offset: int = 0):
        fh = fh
        if offset:
            fh.seek(offset)
        self.threat_id = U64Entry(fh).as_u64
        self.detection_id = UuidEntry(fh).as_uuid
        magic_or_threat_name = WStrEntry(fh).as_str
        # Magic version is missing in some old samples

        if magic_or_threat_name.lower().startswith("magic.version"):
            self.magic_version = magic_or_threat_name
            self.threat_name = WStrEntry(fh).as_str
        else:
            logger.info(
                "No magic in header",
            )
            self.threat_name = magic_or_threat_name
            self.magic_version = None


class WDDHFlags(WDDHPart):
    def __init__(self, fh: BinaryIO, offset: int = 0):
        fh = fh
        if offset:
            fh.seek(offset)

        # This flag is used to assert present of a section
        self.flag_1 = U32Entry(fh).as_u32
        self.flag_2 = U32Entry(fh).as_u32
        self.flag_3 = U32Entry(fh).as_u32
        self.flag_4 = U32Entry(fh).as_u32
        self.flag_5 = U32Entry(fh).as_u32
        self.threat_status_id = ThreatStatusID(U32Entry(fh).as_u32)
        self.flag_list_len = U32Entry(fh).as_u32
        self.flag_list = []
        for i in range(self.flag_list_len):
            self.flag_list.append(U32Entry(fh).as_u32)
        self.alert_detail_count = U32Entry(fh).as_u32


class WDDHInformation(WDDHPart):
    def __init__(self, fh: BinaryIO, offset: int = 0):
        fh = fh
        if offset:
            fh.seek(offset)
        self.magic_version = WStrEntry(fh).as_str
        self.ressource_type = WStrEntry(fh).as_str
        self.ressource_location = WStrEntry(fh).as_str
        self.flag_1 = U32Entry(fh).as_u32
        self.blob_len = U32Entry(fh).as_u32
        self.blob = BlobEntry(fh)


class WDDHMetadata(WDDHPart):
    def __init__(self, fh: BinaryIO, offset: int = 0):
        fh = fh
        if offset:
            fh.seek(offset)
        self.last_threat_status_change = FiletimeEntry(fh)
        self.threat_status_error_code = S32Entry(fh).as_s32
        self.flag_1 = U32Entry(fh).as_u32
        self.unknown_uid = UuidEntry(fh).as_uuid
        self.current_threat_execution_id = U32Entry(fh).as_u32


class WDDHOptionalSection(WDDHPart):
    def __init__(self, fh: BinaryIO, offset: int = 0):
        fh = fh
        if offset:
            fh.seek(offset)
        raise NotImplementedError()


class WDDHMetadata2(WDDHPart):
    def __init__(self, fh: BinaryIO, offset: int = 0):
        if offset:
            fh.seek(offset)
        self.flag_1 = U32Entry(fh).as_u32
        self.domain_user = WStrEntry(fh).as_str
        self.flag_2 = U32Entry(fh).as_u32
        self.process_name = WStrEntry(fh).as_str
        self.flag_3 = U32Entry(fh).as_u32
        self.flag_4 = U32Entry(fh).as_u32
        self.flag_5 = U32Entry(fh).as_u32
        self.initial_detection_time = FiletimeEntry(fh)
        self.flag_6 = U32Entry(fh).as_u32
        self.remediation = FiletimeEntry(fh)
        self.flag_7 = U32Entry(fh).as_u32
        self.unknown_1 = ProbablyBoolEntry(fh)
        self.flag_8 = U32Entry(fh).as_u32
        self.domain_user_group = WStrEntry(fh).as_str  # unsure
        self.flag_9 = U32Entry(fh).as_u32
        self.count_following_information_section = U32Entry(fh).as_u32


class WDDHFooter(WDDHPart):
    def __init__(self, fh: BinaryIO, offset: int = 0):
        if offset:
            fh.seek(offset)
        self.unknown_1 = ProbablyBoolEntry(fh)
        self.flag_1 = U32Entry(fh).as_u32
        self.flag_2 = U32Entry(fh).as_u32
        self.flag_3 = U32Entry(fh).as_u32
        self.flag_4 = U32Entry(fh).as_u32



#!/usr/bin/env python3
import argparse
import collections
import io
import json
import logging
import sys
from importlib.metadata import version
from pathlib import Path
from typing import BinaryIO

from wddh.wddh_clean import WDDHClean

logger = logging.getLogger(__name__)


def gen_argparse() -> argparse.ArgumentParser:
    """

    :return:
    """
    parser = argparse.ArgumentParser(
        description="Parser for Windows Defender Detection history artifact (files located under \ProgramData\Microsoft\Windows Defender\Scans\History\Service\DetectionHistory\)"
    )
    # * pour plusieur fichier
    parser.add_argument(
        "-i",
        "--in",
        required=False,
        type=argparse.FileType("rb"),
        help="Input file",
        dest="infile",
    )
    parser.add_argument(
        "-D",
        "--directory",
        required=False,
        type=Path,
        help="Input directory",
        dest="directory",
    )
    parser.add_argument(
        "-s", "--short", action="store_true", help="Only return a subset of information"
    )
    parser.add_argument(
        "-o",
        "--out",
        nargs="?",
        type=argparse.FileType("w"),
        dest="outfile",
        default=sys.stdout,
    )
    parser.add_argument(
        "-d",
        "--debug",
        help="Logs in debug mode (DEBUG)",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
        default=logging.WARNING,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Logs in verbose mode (INFO)",
        action="store_const",
        dest="loglevel",
        const=logging.INFO,
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s  {version('wddh')}"
    )
    return parser


def custom_json(obj):
    if isinstance(obj, collections.abc.Mapping):
        return dict(obj)
    return str(obj)


def process_one_file(infile: BinaryIO):
    wddh = WDDHClean(infile)
    current_offset = infile.tell()
    infile.seek(0, io.SEEK_END)
    eof = infile.tell()
    infile.close()
    if eof != current_offset:
        logger.warning(
            "File not fully processed : %d read out of %d", current_offset, eof
        )
    return wddh


def dump_full(x: WDDHClean, outfile):
    json.dump(dict(x), outfile, default=custom_json)


def dump_short(x: WDDHClean, outfile):
    json.dump(x.as_short(), outfile, default=custom_json)


def main() -> int:
    args = gen_argparse().parse_args()
    logging.basicConfig(
        level=args.loglevel,
        format="[%(relativeCreated)dms]%(funcName)s:%(lineno)d | %(levelname)s - %(message)s",
    )

    infile = args.infile
    directory: Path = args.directory
    dump_func = dump_full
    if args.short:
        dump_func = dump_short
    if infile:
        dump_func(process_one_file(infile), args.outfile)
    if directory:
        ret = []
        if not directory.is_dir():
            logger.error("%s is not a directory, exit", directory)
            return 1
        for file in directory.rglob("*"):
            if file.is_file():
                with file.open("rb") as f:
                    try:
                        logger.debug("Processing file : %s", file)
                        ret.append(process_one_file(f))
                    except Exception as exc:
                        logger.warning("Fail to parse %s : %s", file, exc)
        ret.sort(key=lambda x: x.metadata_2.initial_detection_time.as_u64)
        for f in ret:
            dump_func(f, args.outfile)
            args.outfile.write("\n")
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
Enum from https://learn.microsoft.com/en-us/previous-versions/windows/desktop/defender/msft-mpthreatdetection



"""

from enum import Enum


class ThreatStatusID(Enum):
    Unknown = 0
    Detected = 1
    Cleaned = 2
    Quarantined = 3
    Removed = 4
    Allowed = 5
    Blocked = 6  # CleanFailed
    QuarantineFailed = 102
    RemoveFailed = 103
    AllowFailed = 104
    Abondoned = 105  # typo from documentation
    BlockedFailed = 107
    MISSING = -1

    @classmethod
    def _missing_(cls, value):
        """
        Set default value to prevent crash is value is missing
        Args:
            value:

        Returns:

        """
        return cls.MISSING
